/*
 * MetodosIterativos.java
 *
 * Creado el 13 de diciembre de 2005, 10:46
 * Modificado el 5 de noviembre de 2007, 19:10
 */


package auxiliar;

/**
 *
 * @author Antonio Pallars, Vctor Jimnez
 *
 * MTODOS NUMRICOS
 * LICENCIATURA DE MATEMTICAS
 * UNIVERSIDAD DE MURCIA
 */
public class MetodosIterativos {
    
    public static double biseccion(Funcion f, double a, double b,
            double precision, int nmaxiter) throws ErrorMetodosIterativos {
        double aa = a;
        double bb = b;
        double h;
        double u = f.eval(aa);
        double v = f.eval(bb);
        double c;
        double w;
        
        if (u * v >= 0) {
            System.out.println("Biseccin: no hay cambio de signo "
                    +"de la funcin.");
            throw new ErrorMetodosIterativos();
        }
        for (int k = 1; k <= nmaxiter; k++){
            h = 0.5 * (bb - aa);
            c = aa + h;
            w = f.eval(c);
            if ((h < precision) || (Math.abs(w) < precision)){
                System.out.println("Raz en c = " + c + "; f(c) = " + w
                        + " en " + k + " pasos.");
                return c;
            }
            if (w * u > 0) {
                aa = c;
                u = w;
            } else {
                bb = c;
                v = w;
            }
        }
        System.out.println("Biseccin: no hay convergencia en "
                + nmaxiter + " iteraciones.");
        throw new ErrorMetodosIterativos();
    }
    
    public static double  secante(Funcion f, double x0, double x1,
            double precision, int nmaxiter) throws ErrorMetodosIterativos {
        double xx0 = x0;
        double xx1 = x1;
        double h;
        double u = f.eval(xx0);
        double v = f.eval(xx1);
        if (Math.abs(u) < precision) {
            System.out.println("Raz en c = " + xx0 + "; f(c) = " + u);
            return xx0;
        }
        if (Math.abs(v) < precision) {
            System.out.println("Raz en c = " + xx1 + "; f(c) = " + v);
            return xx1;
        }
        if (Math.abs(xx0 - xx1) < precision) {
            System.out.println(
                    "Secante: error en las condiciones iniciales.");
            throw new ErrorMetodosIterativos();
        }
        for (int k = 1; k <= nmaxiter; k++) {
            if (Math.abs(u - v) < precision) {
                System.out.println(
                        "Secante: secante horizontal en el paso " + k + ".");
                throw new ErrorMetodosIterativos();
            }
            h = - v * (xx1 - xx0) / (v - u);
            xx0 = xx1;
            u = v;
            xx1 = xx1 + h;
            v = f.eval(xx1);
            if (Math.abs(v) < precision) {
                System.out.println("Raz en c = " + xx1 +
                        "; f(c) = " + v + " en " + k + " pasos.");
                return xx1;
            }
        }
        System.out.println("Secante: no hay convergencia en "
                + nmaxiter + " iteraciones.");
        throw new ErrorMetodosIterativos();
    }
    
    public static double regulaFalsi(Funcion f, double a, double b,
            double precision, int nmaxiter) throws ErrorMetodosIterativos {
        double aa = a;
        double bb = b;
        double h;
        double u = f.eval(aa);
        double v = f.eval(bb);
        double c;
        double w;
        
        if (u * v >= 0){
            System.out.println("Regula Falsi: no hay cambio de signo "
                    + "de la funcin.");
            throw new ErrorMetodosIterativos();
        }
        for (int k = 1; k <= nmaxiter; k++) {
            h = - u * (bb - aa) / (v - u);
            c = aa + h;
            w = f.eval(c);
            if (Math.abs(w) < precision) {
                System.out.println("Raz en c = " + c + "; f(c) = " + w
                        + " en " + k + " pasos.");
                return c;
            }
            if (w * u > 0) {
                aa = c;
                u = w;
            } else {
                bb = c;
                v = w;
            }
        }
        System.out.println("Regula Falsi: no hay convergencia en "
                + nmaxiter + " iteraciones.");
        throw new ErrorMetodosIterativos();
    }
    
    public static double  regulaFalsiModificado(Funcion f, double a, double b,
            double precision, int nmaxiter) throws ErrorMetodosIterativos {
        double aa = a;
        double bb = b;
        double h;
        double u = f.eval(aa);
        double v = f.eval(bb);
        
        /**
         * Modificacin: guardamos memoria del valor de f en el ltimo punto
         * de corte, empezando con aa
         */
        double w1 = u;
        double c2;
        double w2;
        
        if(u * v >= 0) {
            System.out.println("Regula Falsi modificado: no hay cambio"
                    + " de signo de la funcin.");
            throw new ErrorMetodosIterativos();
        }
        for (int k = 1; k <= nmaxiter; k++) {
            h = - u * (bb - aa) / (v - u);
            c2 = aa + h;
            w2 = f.eval(c2);
            if (Math.abs(w2) < precision) {
                System.out.println("Raz en c = " + c2 + "; f(c) = " + w2
                        + " en " + k + " pasos.");
                return c2;
            }
            if (w2 * u > 0) {
                aa = c2;
                u = w2;
                if (w1 * w2 > 0) {
                    v = 0.5 * v;
                }
                w1 = w2;
            } else {
                bb = c2;
                v = w2;
                if (w1 * w2 > 0){
                    u = 0.5 * u;
                }
                w1 = w2;
            }
        }
        System.out.println("Regula Falsi modificado: no hay convergencia en "
                + nmaxiter + " iteraciones.");
        throw new ErrorMetodosIterativos();
    }
    
    public static double newton(Funcion f, double x, double precision,
            int nmaxiter) throws ErrorMetodosIterativos {
        double xx = x;
        double h;
        double u = f.eval(xx);
        double du = f.derivada(xx);
        
        for (int k = 1; k <= nmaxiter; k++) {
            if (Math.abs(u) < precision) {
                System.out.println("Raz en c = " + xx + "; f(c) = " + u +
                        " en " + k + " pasos.");
                return xx;
            }
            if (Math.abs(du) < precision) {
                System.out.println("Newton: derivada nula en el punto.");
                throw new ErrorMetodosIterativos();
            }
            h = - u / du;
            xx = xx + h;
            u = f.eval(xx);
            du = f.derivada(xx);
        }
        System.out.println("Newton: no hay convergencia en " + nmaxiter +
                " iteraciones.");
        throw new ErrorMetodosIterativos();
    }

    public static double iteracionPF(Funcion f, double x, double precision,
            int nmaxiter) throws ErrorMetodosIterativos {
        double xx = x;
        double fx = f.eval(xx);
        
        for (int k = 1; k <= nmaxiter; k++) {
            if (Math.abs(fx - xx) < precision) {
                System.out.println("Punto fijo en c = " + xx + "; f(c) = " + fx
                        + " en " + k + " pasos.");
                return xx;
            }
            xx = fx;
            fx = f.eval(xx);
        }
        System.out.println("Punto fijo: No hay convergencia en " + nmaxiter
                + " iteraciones.");
        throw new ErrorMetodosIterativos();
    }
    
    public static double iteracionPFacelerada(Funcion f, double x,
            double precision, int nmaxiter) throws ErrorMetodosIterativos {
        double xx = x;
        double fx = f.eval(xx);
        double ffx;
        double xxacelerado;
        
        for(int k = 1; k <= nmaxiter; k++) {
            if(Math.abs(fx - xx) < precision) {
                System.out.println("Punto fijo en c = " + xx + "; f(c) = " 
                        + fx + " en " + k + " pasos.");
                return xx;
            }
            ffx = f.eval(fx);
            xxacelerado = xx - ((fx - xx) * (fx - xx)) / (ffx - 2 * fx + xx);
            if(Math.abs(f.eval(xxacelerado) - xxacelerado) < precision) {
                System.out.println("Punto fijo en c = " + xxacelerado +
                        "; f(c) = " + f.eval(xxacelerado) + " en " + k + 
                        " pasos.");
                return xxacelerado;
            }
            xx = fx;
            fx = f.eval(xx);
        }
        System.out.println("Punto fijo acelerado: No hay convergencia en " 
                + nmaxiter + " iteraciones.");        
        throw new ErrorMetodosIterativos();
    }
    
    public static double steffensenPF(Funcion f, double x, double precision,
            int nmaxiter) throws ErrorMetodosIterativos {
        double xx = x;
        double fx = f.eval(xx);
        double ffx = f.eval(fx);
        
        for (int k = 1; k <= nmaxiter; k++) {
            if (Math.abs(fx - xx) < precision) {
                System.out.println("Punto fijo en c = " + xx + "; f(c) = " + fx
                        + " en " + k + " pasos.");
                return xx;
            }
            xx = xx - ((fx - xx) * (fx - xx)) / (ffx - 2 * fx + xx);
            fx = f.eval(xx);
            ffx = f.eval(fx);
        }
        System.out.println("Steffensen: no hay convergencia en " + nmaxiter
                + " iteraciones.");
        throw new ErrorMetodosIterativos();
    }
        
    /**
     * Mtodo de Newton para races multiples sin conocer
     * su multiplicidad. Consiste en aplicar el mtodo de Newton convencional
     * a la funcin F(x) = f(x)/f'(x), que en el caso de que c sea una raz
     * mltiple de f, tiene a c como una raz simple. 
     *
     * En este caso, la funcin g(x) = x - f(x)/f'(x) que define habitualmente
     * el mtodo de Newton se convierte en 
     *   G(x) = x - F(x)/F'(x) = x - f(x)f'(x)/((f'(x))^2 - f(x)f''(x)).     
     */
    public static double  newtonMultiple(FuncionC2 f, double x,
            double precision, int nmaxiter) throws ErrorMetodosIterativos {
        double xx = x;
        double h;
        double u = f.eval(xx);
        double du = f.derivada(xx);
        double d2u = f.derivada2(xx);
        
        for (int k = 1; k <= nmaxiter; k++) {
            if (Math.abs(u) < precision) {
                System.out.println("Raz en c = " + xx + "; f(c) = " + u
                        + " en "+ k + " pasos.");
                return xx;
            }
            if (Math.abs(du * du - u * d2u) < precision) {
                System.out.println("Newton para races mltiples: "
                        + "denominador nulo en el punto.");
                throw new ErrorMetodosIterativos();
            }
            h = - (u * du) / (du * du - u * d2u);
            xx = xx + h;
            u = f.eval(xx);
            du = f.derivada(xx);
            d2u = f.derivada2(xx);
        }
        System.out.println("Newton para races mltiples: "
                + "no hay convergencia en " + nmaxiter + " iteraciones.");
        throw new ErrorMetodosIterativos();
    }
    
    /**
     * Mejora del mtodo de Newton para races multiples sin conocer
     * su multiplicidad que tambin proporciona el nmero de iteraciones
     * realizadas
     */
    public static double[]  newtonMultipleMejorado(FuncionC2 f, double x,
            double precision, int nmaxiter) throws ErrorMetodosIterativos {
        double xx = x;
        double h;
        double u = f.eval(xx);
        double du = f.derivada(xx);
        double d2u = f.derivada2(xx);
        double[] resultado = new double[2];
        
        for (int k = 1; k <= nmaxiter; k++) {
            if (Math.abs(u) < precision) {
                System.out.println("Raz en c = " + xx + "; f(c) = " + u
                        + " en "+ k + " pasos.");
                resultado[0] = xx;
                resultado[1] = k;
                return resultado;
            }
            if (Math.abs(du * du - u * d2u) < precision) {
                System.out.println("Newton para races mltiples: "
                        + "denominador nulo en el punto.");
                throw new ErrorMetodosIterativos();
            }
            h = - (u * du) / (du * du - u * d2u);
            xx = xx + h;
            u = f.eval(xx);
            du = f.derivada(xx);
            d2u = f.derivada2(xx);
        }
        System.out.println("Newton para races mltiples: "
                + "no hay convergencia en " + nmaxiter + " iteraciones.");
        throw new ErrorMetodosIterativos();
    }

    /**
     * Excepcin ErrorMetodosIterativos para ser devuelta por 
     * los mtodos de la clase
     */
    public static class  ErrorMetodosIterativos extends Exception {
        ErrorMetodosIterativos() {
            super("Error al aproximar races");
        }
    }
    
}

